/* bobscn.c - a lexical scanner */
/*
	Copyright (c) 1991, by David Michael Betz
	All rights reserved
*/

#include <setjmp.h>
#include "bob.h"

/* useful definitions */
#define LSIZE	300

/* keyword table */
static struct { char *kt_keyword; int kt_token; } ktab[] = {
{ "class",	T_CLASS		},
{ "static",	T_STATIC	},
{ "if",		T_IF		},
{ "else",	T_ELSE		},
{ "while",	T_WHILE		},
{ "return",	T_RETURN	},
{ "for",	T_FOR		},
{ "break",	T_BREAK		},
{ "continue", 	T_CONTINUE	},
{ "do",		T_DO		},
{ "new",	T_NEW		},
{ "delete",	T_DELETE	},
{ "nil",	T_NIL		},
{ "null",   T_NIL       },
{ NULL,		0		}};

/* token name table */
static char *t_names[] = {
"delete",
"%=",
"<float>",
"<notoken>",
"<eof>",
"<string>",
"<ident>",
"<long>",
"class",
"static",
"if",
"else",
"while",
"return",
"for",
"break",
"continue",
"do",
"new",
"null",
"<=",
"==",
"!=",
">=",
"<<",
">>",
"&&",
"||",
"++",
"--",
"+=",
"-=",
"*=",
"/=",
"::",
"->"
};

/* global variables */
long t_value;		/* numeric value */
char t_token[TKNSIZE+1];/* token string */

/* local variables */
static int (*getcf)();	/* getc function */
static void *getcd;	/* getc data */
static int savetkn;	/* look ahead token */
static int savech;	/* look ahead character */
static int lastch;	/* last input character */
static char line[LSIZE];/* last input line */
static char *lptr;	/* line pointer */
static int lnum;	/* line number */

/* init_scanner - initialize the scanner */
init_scanner(gf,gd)
  int (*gf)(); void *gd;
{
    /* remember the getc function and data */
    getcf = gf; getcd = gd;

    /* setup the line buffer */
    lptr = line; *lptr = '\0';
    lnum = 0;

    /* no lookahead yet */
    savetkn = T_NOTOKEN;
    savech = '\0';

    /* no last character */
    lastch = '\0';
}

/* token - get the next token */
int token()
{
    int tkn;

    if ((tkn = savetkn) != T_NOTOKEN)
	savetkn = T_NOTOKEN;
    else
	tkn = rtoken();
    return (tkn);
}

/* stoken - save a token */
stoken(tkn)
  int tkn;
{
    savetkn = tkn;
}

/* tkn_name - get the name of a token */
char *tkn_name(tkn)
  int tkn;
{
    static char tname[2];
    if (tkn <= _TMAX)
	return (t_names[tkn-_TMIN]);
    tname[0] = tkn;
    tname[1] = '\0';
    return (tname);
}

/* rtoken - read the next token */
static int rtoken()
{
    int ch,ch2;

    /* check the next character */
    for (;;)
	switch (ch = skipspaces()) {
	case EOF:	return (T_EOF);
	case '"':	return (getstring());
	case '\'':	return (getcharacter());
	case '<':	switch (ch = getch()) {
			case '=':
			    return (T_LE);
			case '<':
			    return (T_SHL);
			default:
			    savech = ch;
			    return ('<');
			}
	case '=':	switch (ch = getch()) {
			case '=':
			    return (T_EQ);
			default:
			    savech = ch;
			    return ('=');
			}
	case '!':	switch (ch = getch()) {
			case '=':
			    return (T_NE);
			default:
			    savech = ch;
			    return ('!');
			}
	case '>':	switch (ch = getch()) {
			case '=':
			    return (T_GE);
			case '>':
			    return (T_SHR);
			default:
			    savech = ch;
			    return ('>');
			}
	case '&':	switch (ch = getch()) {
			case '&':
			    return (T_AND);
			default:
			    savech = ch;
			    return ('&');
			}
	case '|':	switch (ch = getch()) {
			case '|':
			    return (T_OR);
			default:
			    savech = ch;
			    return ('|');
			}
	case '+':	switch (ch = getch()) {
			case '+':
			    return (T_INC);
			case '=':
			    return (T_ADDEQ);
			default:
			    savech = ch;
			    return ('+');
			}
	case '-':	switch (ch = getch()) {
			case '-':
			    return (T_DEC);
			case '=':
			    return (T_SUBEQ);
			case '>':
			    return (T_MEMREF);
			default:
			    savech = ch;
			    return ('-');
			}
	case '*':	switch (ch = getch()) {
			case '=':
			    return (T_MULEQ);
			default:
			    savech = ch;
			    return ('*');
			}
    case '%':	switch (ch = getch()) {
			case '=':
			    return (T_REMEQ);
			default:
			    savech = ch;
			    return ('%');
			}
	case '@':
			while ((ch = getch()) != EOF)
			if (ch == '\n')
				break;
			break;
	case '/':	switch (ch = getch()) {
			case '=':
			    return (T_DIVEQ);
			case '/':
			    while ((ch = getch()) != EOF)
				if (ch == '\n')
				    break;
			    break;
			case '*':
			    ch = ch2 = EOF;
			    for (; (ch2 = getch()) != EOF; ch = ch2)
				if (ch == '*' && ch2 == '/')
				    break;
			    break;
			default:
			    savech = ch;
			    return ('/');
			}
			break;
	case ':':	switch (ch = getch()) {
			case ':':
			    return (T_CC);
			default:
			    savech = ch;
			    return (':');
			}
	default:	if (isdigit(ch))
			    return (getnumber(ch));
			else if (isidchar(ch))
			    return (getid(ch));
			else {
			    t_token[0] = ch;
			    t_token[1] = '\0';
			    return (ch);
			}
	}
}

/* getstring - get a string */
static int getstring()
{
    // char *p;
    int i,ch;

    /* get the string */
    // p = t_token;
    i=0;
    while ((ch = literalch()) != EOF && ch != '"')
    {
		if (i >= TKNSIZE)
		   parse_error("string literal too long.");
		t_token[i] = ch;
		i++;
	}
	//*p++ = ch;
    if (ch == EOF)
	savech = EOF;
    //*p = '\0';
    t_token[i] = '\0';
    return (T_STRING);
}

/* getcharacter - get a character constant */
static int getcharacter()
{
    t_value = literalch();
    t_token[0] = t_value;
    t_token[1] = '\0';
    if (getch() != '\'')
	parse_error("closing single quote expected");
    return (T_NUMBER);
}

/* literalch - get a character from a literal string */
static int literalch()
{
    int ch;
    if ((ch = getch()) == '\\')
	switch (ch = getch()) {
	case 'n':  ch = '\n'; break;
	case 'r':  ch = '\r'; break;
	case 't':  ch = '\t'; break;
	case 'v':  ch = '\v'; break;
	case 'f':  ch = '\f'; break;
	case 'a':  ch = '\a'; break;
	case 'b':  ch = '\b'; break;
	case '0':  ch = '\0'; break;
	case EOF:  ch = '\\'; savech = EOF; break;
	}
    return (ch);
}

/* getid - get an identifier */
static int getid(ch)
  int ch;
{
    char *p;
    int i;

    /* get the identifier */
    p = t_token; *p++ = ch;
    while ((ch = getch()) != EOF && isidchar(ch))
	*p++ = ch;
    savech = ch;
    *p = '\0';

    /* check to see if it is a keyword */
    for (i = 0; ktab[i].kt_keyword != NULL; ++i)
	if (strcmp(ktab[i].kt_keyword,t_token) == 0)
	    return (ktab[i].kt_token);
    return (T_IDENTIFIER);
}

/* getnumber - get a number */
static int getnumber(ch)
  int ch;
{
    char *p;
    int xnum = 0;
    int ttype = T_NUMBER;
    /* get the number */

    p = t_token; *p++ = ch; t_value = 0;

    while ((ch = getch()) != EOF)
    {
		if (!isdigit(ch))
		{
			if (!xnum)
			{
#ifdef USE_FLOAT
    			 if (ch == '.')
				 {
					 if (ttype == T_FNUMBER)
					    break;
				      ttype = T_FNUMBER;
				  }
				  else if ((ch == 'E') || (ch == 'e'))
				  {
					  ttype = T_FNUMBER;
					  *p++ = ch;
					  ch = getch();
					  if (!(isdigit(ch) || (ch == '-') || (ch == '+')))
					    break;
				  }
				   else
#endif
				   if  ((*(p-1)=='0') && (tolower(ch) == 'x'))
				     xnum = 1;
					else
					  break;
			}
			else if (!isxdigit(ch))
			   break;
	    }
	    *p++ = ch;
    }
    savech = ch;
    *p = '\0';

#ifdef USE_FLOAT
    if (ttype == T_NUMBER)
       t_value = strtol(t_token,0,0);
    else
    {
       float fv;
       float* f = &t_value;
       sscanf(t_token,"%f",&fv);
       *f = fv;
	}
#else
    t_value = strtol(t_token,0,0);
#endif
    return (ttype);
}

/* skipspaces - skip leading spaces */
static skipspaces()
{
    int ch;
    while ((ch = getch()) != '\0' && isspace(ch))
	;
    return (ch);
}

/* isidchar - is this an identifier character */
static int isidchar(ch)
  int ch;
{
    return (isupper(ch)
         || islower(ch)
         || isdigit(ch)
         || ch == '_');
}

/* getch - get the next character */
static int getch()
{
    int ch;

    /* check for a lookahead character */
    if ((ch = savech) != '\0')
	savech = '\0';

    /* check for a buffered character */
    else {
	while ((ch = *lptr++) == '\0') {

	    /* check for being at the end of file */
	    if (lastch == EOF)
		return (EOF);

	    /* read the next line */
	    lptr = line;
	    while ((lastch = (*getcf)(getcd)) != EOF && lastch != '\n')
		*lptr++ = lastch;
	    *lptr++ = '\n'; *lptr = '\0';
	    lptr = line;
	    ++lnum;
	}
    }

    /* return the current character */
    return (ch);
}

/* parse_error - report an error in the current line */
parse_error(msg)
  char *msg;
{
    extern jmp_buf error_trap;
    char buf[LSIZE],*src,*dst;

    /* redisplay the line with the error */
    sprintf(buf,"--> %s \nin line %d \n%s",msg,lnum,line);
    osputs(buf);

    /* point to the position immediately following the error */
    for (src = line, dst = buf; src < lptr-1; ++src)
	*dst++ = (*src == '\t' ? '\t' : ' ');
    *dst++ = '^'; *dst++ = '\n'; *dst = '\0';
    osputs(buf);

    /* invoke the error trap */
    longjmp(error_trap,1);
}
